# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from datetime import timezone

import pytest

from google.protobuf import timestamp_pb2

import proto


def test_repeated_composite_init():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.RepeatedField(proto.MESSAGE, message=Foo, number=1)

    baz = Baz(foos=[Foo(bar=42)])
    assert len(baz.foos) == 1
    assert baz.foos[0].bar == 42


def test_repeated_composite_init_struct():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.RepeatedField(proto.MESSAGE, message=Foo, number=1)

    baz = Baz(foos=[{'bar': 42}])
    assert len(baz.foos) == 1
    assert baz.foos[0].bar == 42


def test_repeated_composite_falsy_behavior():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.RepeatedField(proto.MESSAGE, message=Foo, number=1)

    baz = Baz()
    assert not baz.foos
    assert len(baz.foos) == 0


def test_repeated_composite_marshaled():
    class Foo(proto.Message):
        timestamps = proto.RepeatedField(proto.MESSAGE,
            message=timestamp_pb2.Timestamp,
            number=1,
        )

    foo = Foo(timestamps=[datetime(2012, 4, 21, 15, tzinfo=timezone.utc)])
    foo.timestamps.append(timestamp_pb2.Timestamp(seconds=86400 * 365))
    foo.timestamps.append(datetime(2017, 10, 14, tzinfo=timezone.utc))
    assert all([isinstance(i, datetime) for i in foo.timestamps])
    assert all([isinstance(i, timestamp_pb2.Timestamp)
                for i in Foo.pb(foo).timestamps])
    assert foo.timestamps[0].year == 2012
    assert foo.timestamps[0].month == 4
    assert foo.timestamps[0].hour == 15
    assert foo.timestamps[1].year == 1971
    assert foo.timestamps[1].month == 1
    assert foo.timestamps[1].hour == 0
    assert foo.timestamps[2].year == 2017
    assert foo.timestamps[2].month == 10
    assert foo.timestamps[2].hour == 0


def test_repeated_composite_outer_write():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.RepeatedField(proto.MESSAGE, message=Foo, number=1)

    baz = Baz()
    baz.foos = [Foo(bar=96), Foo(bar=48)]
    assert len(baz.foos) == 2
    assert baz.foos[0].bar == 96
    assert baz.foos[1].bar == 48


def test_repeated_composite_append():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.RepeatedField(proto.MESSAGE, message=Foo, number=1)

    baz = Baz()
    baz.foos.append(Foo(bar=96))
    baz.foos.append({'bar': 72})
    assert len(baz.foos) == 2
    assert baz.foos[0].bar == 96
    assert baz.foos[1].bar == 72


def test_repeated_composite_iadd():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.RepeatedField(proto.MESSAGE, message=Foo, number=1)

    baz = Baz()
    baz.foos += [Foo(bar=96), Foo(bar=48)]
    assert len(baz.foos) == 2
    assert baz.foos[0].bar == 96
    assert baz.foos[1].bar == 48


def test_repeated_composite_set():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.RepeatedField(proto.MESSAGE, message=Foo, number=1)

    baz = Baz(foos=[{'bar': 96}, {'bar': 48}])
    baz.foos[1] = Foo(bar=55)
    assert baz.foos[0].bar == 96
    assert baz.foos[1].bar == 55


def test_repeated_composite_set_wrong_type():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class NotFoo(proto.Message):
        eggs = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.RepeatedField(proto.MESSAGE, message=Foo, number=1)

    baz = Baz()
    with pytest.raises(TypeError):
        baz.foos.append(NotFoo(eggs=42))



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import pytest

import proto


def test_repeated_scalar_init():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    assert foo.bar == [1, 1, 2, 3, 5, 8, 13]


def test_repeated_scalar_append():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    foo.bar.append(21)
    foo.bar.append(34)
    assert foo.bar == [1, 1, 2, 3, 5, 8, 13, 21, 34]


def test_repeated_scalar_iadd():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    foo.bar += [21, 34]
    assert foo.bar == [1, 1, 2, 3, 5, 8, 13, 21, 34]


def test_repeated_scalar_setitem():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    foo.bar[4] = 99
    assert foo.bar == [1, 1, 2, 3, 99, 8, 13]
    assert foo.bar[4] == 99


def test_repeated_scalar_overwrite():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    foo.bar = [1, 2, 4, 8, 16]
    assert foo.bar == [1, 2, 4, 8, 16]


def test_repeated_scalar_eq_ne():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    assert foo.bar == copy.copy(foo.bar)
    assert foo.bar != [1, 2, 4, 8, 16]


def test_repeated_scalar_del():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    del foo.bar
    assert foo.bar == []
    assert not foo.bar


def test_repeated_scalar_delitem():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    del foo.bar[5]
    del foo.bar[3]
    assert foo.bar == [1, 1, 2, 5, 13]


def test_repeated_scalar_sort():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[8, 1, 2, 1, 21, 3, 13, 5])
    foo.bar.sort()
    assert foo.bar == [1, 1, 2, 3, 5, 8, 13, 21]


def test_repeated_scalar_wrong_type():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT32, number=1)

    foo = Foo(bar=[1, 1, 2, 3, 5, 8, 13])
    with pytest.raises(TypeError):
        foo.bar.append(21.0)
    with pytest.raises(TypeError):
        foo.bar.append('21')



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from google.protobuf import descriptor_pool
from google.protobuf import symbol_database

from proto import message


def pytest_runtest_setup(item):
    pool = descriptor_pool.DescriptorPool()
    sym_db = symbol_database.SymbolDatabase(pool=pool)
    message.registry.clear()
    item._mocks = (
        mock.patch.object(descriptor_pool, 'Default', return_value=pool),
        mock.patch.object(symbol_database, 'Default', return_value=sym_db),
    )
    [i.start() for i in item._mocks]


def pytest_runtest_teardown(item):
    [i.stop() for i in item._mocks]



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fields import Field
from .fields import MapField
from .fields import RepeatedField
from .marshal import marshal
from .message import Message
from .primitives import ProtoType


DOUBLE = ProtoType.DOUBLE
FLOAT = ProtoType.FLOAT
INT64 = ProtoType.INT64
UINT64 = ProtoType.UINT64
INT32 = ProtoType.INT32
FIXED64 = ProtoType.FIXED64
FIXED32 = ProtoType.FIXED32
BOOL = ProtoType.BOOL
STRING = ProtoType.STRING
MESSAGE = ProtoType.MESSAGE
BYTES = ProtoType.BYTES
UINT32 = ProtoType.UINT32
ENUM = ProtoType.ENUM
SFIXED32 = ProtoType.SFIXED32
SFIXED64 = ProtoType.SFIXED64
SINT32 = ProtoType.SINT32
SINT64 = ProtoType.SINT64


__all__ = (
    'Field',
    'MapField',
    'RepeatedField',
    'marshal',
    'Message',

    # Expose the types directly.
    'DOUBLE',
    'FLOAT',
    'INT64',
    'UINT64',
    'INT32',
    'FIXED64',
    'FIXED32',
    'BOOL',
    'STRING',
    'MESSAGE',
    'BYTES',
    'UINT32',
    'ENUM',
    'SFIXED32',
    'SFIXED64',
    'SINT32',
    'SINT64',
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import collections
import copy

from google.protobuf import message
from google.protobuf import duration_pb2
from google.protobuf import timestamp_pb2
from google.protobuf import wrappers_pb2
from google.protobuf.internal import containers

from proto.marshal.types import dates
from proto.marshal.types import wrappers


class Rule(abc.ABC):
    """Abstract class definition for marshal rules."""

    @classmethod
    def __subclasshook__(cls, C):
        if hasattr(C, 'to_python') and hasattr(C, 'to_proto'):
            return True
        return NotImplemented


class MarshalRegistry:
    """A class to translate between protocol buffers and Python classes.

    Protocol buffers defines many common types (e.g. Timestamp, Duration)
    which also exist in the Python standard library. The marshal essentially
    translates between these: it keeps a registry of common protocol buffers
    and their Python representations, and translates back and forth.

    The protocol buffer class is always the "key" in this relationship; when
    presenting a message, the declared field types are used to determine
    whether a value should be transformed into another class. Similarly,
    when accepting a Python value (when setting a field, for example),
    the declared field type is still used. This means that, if appropriate,
    multiple protocol buffer types may use the same Python type.

    The marshal is intended to be a singleton; this module instantiates
    and exports one marshal, which is imported throughout the rest of this
    library. This allows for an advanced case where user code registers
    additional types to be marshaled.
    """
    def __init__(self):
        self._registry = {}
        self._noop = NoopMarshal()

        # Register date and time wrappers.
        self.register(timestamp_pb2.Timestamp, dates.TimestampMarshal())
        self.register(duration_pb2.Duration, dates.DurationMarshal())

        # Register nullable primitive wrappers.
        self.register(wrappers_pb2.BoolValue, wrappers.BoolValueMarshal())
        self.register(wrappers_pb2.BytesValue, wrappers.BytesValueMarshal())
        self.register(wrappers_pb2.DoubleValue, wrappers.DoubleValueMarshal())
        self.register(wrappers_pb2.FloatValue, wrappers.FloatValueMarshal())
        self.register(wrappers_pb2.Int32Value, wrappers.Int32ValueMarshal())
        self.register(wrappers_pb2.Int64Value, wrappers.Int64ValueMarshal())
        self.register(wrappers_pb2.StringValue, wrappers.StringValueMarshal())
        self.register(wrappers_pb2.UInt32Value, wrappers.UInt32ValueMarshal())
        self.register(wrappers_pb2.UInt64Value, wrappers.UInt64ValueMarshal())

    def register(self, proto_type: type, rule: Rule = None):
        """Register a rule against the given ``proto_type``.

        This function expects a ``proto_type`` (the descriptor class) and
        a ``rule``; an object with a ``to_python`` and ``to_proto`` method.
        Each method should return the appropriate Python or protocol buffer
        type, and be idempotent (e.g. accept either type as input).

        This function can also be used as a decorator::

            @marshal.register(timestamp_pb2.Timestamp)
            class TimestampMarshal:
                ...

        In this case, the class will be initialized for you with zero
        arguments.

        Args:
            proto_type (type): A protocol buffer message type.
            rule: A marshal object
        """
        # Sanity check: Do not register anything to a class that is not
        # a protocol buffer message.
        if not issubclass(proto_type, message.Message):
            raise TypeError('Only protocol buffer messages may be registered '
                            'to the marshal.')

        # If a rule was provided, register it and be done.
        if rule:
            # Ensure the rule implements Rule.
            if not isinstance(rule, Rule):
                raise TypeError('Marshal rule instances must implement '
                                '`to_proto` and `to_python` methods.')

            # Register the rule.
            self._registry[proto_type] = rule
            return

        # Create an inner function that will register an instance of the
        # marshal class to this object's registry, and return it.
        def register_rule_class(rule_class: type):
            # Ensure the rule class is a valid rule.
            if not issubclass(rule_class, Rule):
                raise TypeError('Marshal rule subclasses must implement '
                                '`to_proto` and `to_python` methods.')

            # Register the rule class.
            self._registry[proto_type] = rule_class()
            return rule_class
        return register_rule_class

    def to_python(self, proto_type, value, *, absent: bool = None):
        # Internal protobuf has its own special type for lists of values.
        # Return a view around it that implements MutableSequence.
        if isinstance(value, containers.RepeatedCompositeFieldContainer):
            return RepeatedComposite(value)
        if isinstance(value, containers.RepeatedScalarFieldContainer):
            return Repeated(value)

        # Same thing for maps of messages.
        if isinstance(value, containers.MessageMap):
            return MapComposite(value)

        # Convert ordinary values.
        rule = self._registry.get(proto_type, self._noop)
        return rule.to_python(value, absent=absent)

    def to_proto(self, proto_type, value, *, strict: bool = False):
        # For our repeated and map view objects, simply return the
        # underlying pb.
        if isinstance(value, (Repeated, MapComposite)):
            return value.pb

        # Convert lists and tuples recursively.
        if isinstance(value, (list, tuple)):
            return type(value)([self.to_proto(proto_type, i) for i in value])

        # Convert dictionaries recursively when the proto type is a map.
        # This is slightly more complicated than converting a list or tuple
        # because we have to step through the magic that protocol buffers does.
        #
        # Essentially, a type of map<string, Foo> will show up here as
        # a FoosEntry with a `key` field, `value` field, and a `map_entry`
        # annotation. We need to do the conversion based on the `value`
        # field's type.
        if isinstance(value, dict) and (proto_type.DESCRIPTOR.has_options and
                proto_type.DESCRIPTOR.GetOptions().map_entry):
            return {k: self.to_proto(type(proto_type().value), v)
                    for k, v in value.items()}

        # Convert ordinary values.
        rule = self._registry.get(proto_type, self._noop)
        pb_value = rule.to_proto(value)

        # Sanity check: If we are in strict mode, did we get the value we want?
        if strict and not isinstance(pb_value, proto_type):
            raise TypeError(
                'Parameter must be instance of the same class; '
                'expected {expected}, got {got}'.format(
                    expected=proto_type.__name__,
                    got=pb_value.__class__.__name__,
                ),
            )

        # Return the final value.
        return pb_value


class Repeated(collections.MutableSequence):
    """A view around a mutable sequence in protocol buffers.

    This implements the full Python MutableSequence interface, but all methods
    modify the underlying field container directly.
    """
    def __init__(self, sequence: containers.BaseContainer):
        self._pb = sequence

    def __copy__(self):
        """Copy this object and return the copy."""
        return type(self)(sequence=copy.copy(self.pb))

    def __delitem__(self, key):
        """Delete the given item."""
        del self.pb[key]

    def __eq__(self, other):
        if hasattr(other, 'pb'):
            return self.pb._values == other.pb._values
        return self.pb._values == other

    def __getitem__(self, key):
        """Return the given item."""
        return self.pb[key]

    def __len__(self):
        """Return the length of the sequence."""
        return len(self.pb)

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return repr(self.pb)

    def __setitem__(self, key, value):
        self.pb[key] = value

    def insert(self, index: int, value):
        """Insert ``value`` in the sequence before ``index``."""
        self.pb.insert(index, value)

    def sort(self, *, key: str = None, reverse: bool = False):
        """Stable sort *IN PLACE*."""
        self.pb.sort(key=key, reverse=reverse)

    @property
    def pb(self):
        return self._pb


class RepeatedComposite(Repeated):
    """A view around a mutable sequence of messages in protocol buffers.

    This implements the full Python MutableSequence interface, but all methods
    modify the underlying field container directly.
    """
    @property
    def _pb_type(self):
        """Return the protocol buffer type for this sequence."""
        return self.pb._message_descriptor._concrete_class

    def __getitem__(self, key):
        return marshal.to_python(self._pb_type, self.pb[key])

    def __setitem__(self, key, value):
        pb_value = marshal.to_proto(self._pb_type, value, strict=True)
        self.pb._values[key] = pb_value
        self.pb._message_listener.Modified()

    def insert(self, index: int, value):
        """Insert ``value`` in the sequence before ``index``."""
        pb_value = marshal.to_proto(self._pb_type, value, strict=True)
        self.pb._values.insert(index, pb_value)
        self.pb._message_listener.Modified()


class MapComposite(collections.MutableMapping):
    """A view around a mutable sequence in protocol buffers.

    This implements the full Python MutableMapping interface, but all methods
    modify the underlying field container directly.
    """
    @property
    def _pb_type(self):
        """Return the protocol buffer type for this sequence."""
        return self.pb._message_descriptor._concrete_class

    def __init__(self, sequence: containers.BaseContainer):
        self._pb = sequence

    def __contains__(self, key):
        # Protocol buffers is so permissive that querying for the existence
        # of a key will in of itself create it.
        #
        # By taking a tuple of the keys and querying that, we avoid sending
        # the lookup to protocol buffers and therefore avoid creating the key.
        return key in tuple(self.keys())

    def __getitem__(self, key):
        return marshal.to_python(self._pb_type, self.pb[key])

    def __setitem__(self, key, value):
        pb_value = marshal.to_proto(self._pb_type, value, strict=True)
        self.pb._values[key] = pb_value
        self.pb._message_listener.Modified()

    def __delitem__(self, key):
        del self.pb._values[key]
        self.pb._message_listener.Modified()

    def __len__(self):
        return len(self.pb)

    def __iter__(self):
        return iter(self.pb)

    @property
    def pb(self):
        return self._pb


class NoopMarshal:
    """A catch-all marshal that does nothing."""

    def to_python(self, pb_value, *, absent: bool = None):
        return pb_value

    def to_proto(self, value):
        return value


marshal = MarshalRegistry()

__all__ = (
    'marshal',
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import collections.abc
import copy
import re
from typing import List, Type

from google.protobuf import descriptor
from google.protobuf.descriptor_pb2 import MessageOptions
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import symbol_database

from proto.fields import Field
from proto.fields import MapField
from proto.fields import RepeatedField
from proto.marshal import marshal
from proto.marshal.types.message import MessageMarshal
from proto.primitives import ProtoType


class MessageMeta(type):
    """A metaclass for building and registering Message subclasses."""

    def __new__(mcls, name, bases, attrs):
        # Do not do any special behavior for Message itself.
        if not bases:
            return super().__new__(mcls, name, bases, attrs)

        # Pop metadata off the attrs.
        Meta = attrs.pop('Meta', object())

        # A package and full name should be present.
        package = getattr(Meta, 'package', '')
        full_name = getattr(Meta, 'full_name',
            '.'.join((package, attrs.get('__qualname__', name))).lstrip('.'),
        )

        # Special case: Maps. Map fields are special; they are essentially
        # shorthand for a nested message and a repeated field of that message.
        # Decompose each map into its constituent form.
        # https://developers.google.com/protocol-buffers/docs/proto3#maps
        for key, field in copy.copy(attrs).items():
            if not isinstance(field, MapField):
                continue

            # Determine the name of the entry message.
            message_name = '{pascal_key}Entry'.format(
                pascal_key=re.sub(
                    r'_[\w]',
                    lambda m: m.group().upper(),
                    key,
                ).capitalize(),
            )

            # Create the "entry" message (with the key and value fields).
            attrs[message_name] = MessageMeta(message_name, (Message,), {
                'key': Field(field.map_key_type, number=1),
                'value': Field(field.proto_type, number=2,
                    enum=field.enum,
                    message=field.message,
                ),
                'Meta': type('Meta', (object,), {
                    'full_name': '{0}.{1}'.format(full_name, message_name),
                    'options': MessageOptions(map_entry=True),
                    'package': package,
                }),
            })

            # Create the repeated field for the entry message.
            attrs[key] = RepeatedField(ProtoType.MESSAGE,
                number=field.number,
                message=attrs[message_name],
            )

        # Okay, now we deal with all the rest of the fields.
        # Iterate over all the attributes and separate the fields into
        # their own sequence.
        fields = []
        oneofs = collections.OrderedDict()
        index = 0
        for key, field in copy.copy(attrs).items():
            # Sanity check: If this is not a field, do nothing.
            if not isinstance(field, Field):
                continue

            # Remove the field from the attrs dictionary; the field objects
            # themselves should not valuebe direct attributes.
            attrs.pop(key)

            # Add data that the field requires that we do not take in the
            # constructor because we can derive it from the metaclass.
            # (The goal is to make the declaration syntax as nice as possible.)
            field.mcls_data = {
                'name': key,
                'full_name': '{0}.{1}'.format(full_name, key),
                'index': index,
            }

            # Add the field to the list of fields.
            fields.append(field)

            # If this field is part of a "oneof", ensure the oneof itself
            # is represented.
            if field.oneof:
                # In theory, most oneofs will be encountered more than once.
                if field.oneof in oneofs:
                    # Since the oneof exists, simply add this field to it.
                    oneofs[field.oneof].fields.append(field.descriptor)
                else:
                    # Create the oneof descriptor, and add it to the running
                    # dictionary of oneofs.
                    oneofs.setdefault(field.oneof, descriptor.OneofDescriptor(
                        containing_type=None,
                        fields=[field.descriptor],
                        full_name='{0}.{1}'.format(full_name, field.oneof),
                        index=len(oneofs),
                        name=field.oneof,
                    ))

                # The field's descriptor itself also must be given a reference
                # to the oneof that contains it.
                field.descriptor.containing_oneof = oneofs[field.oneof]

            # Increment the field index counter.
            index += 1

        # Get a file descriptor object.
        module = attrs.get('__module__', name.lower()).replace('.', '/')
        if module not in _file_descriptor_registry:
            _file_descriptor_registry[module] = descriptor.FileDescriptor(
                name='%s.proto' % module,
                package=package,
                syntax='proto3',
            )

        # Retrieve any message options.
        opts = getattr(Meta, 'options', MessageOptions())

        # Create the underlying proto descriptor.
        # This programatically duplicates the default code generated
        # by protoc.
        desc = descriptor.Descriptor(
            name=name, full_name=full_name,
            file=_file_descriptor_registry[module],
            filename=None, containing_type=None,
            fields=[i.descriptor for i in fields],
            nested_types=[], enum_types=[], extensions=[],
            oneofs=[i for i in oneofs.values()],
            serialized_options=opts.SerializeToString(),
            syntax='proto3',
        )
        _file_descriptor_registry[module].message_types_by_name[name] = desc

        # Create the MessageInfo instance to be attached to this message.
        attrs['_meta'] = MessageInfo(
            descriptor=desc,
            fields=fields,
            full_name=full_name,
            options=opts,
            package=package,
        )

        # Run the superclass constructor.
        cls = super().__new__(mcls, name, bases, attrs)

        # The info class and fields need a reference to the class just created.
        cls._meta.parent = cls
        for field in cls._meta.fields.values():
            field.parent = cls

        # Attempt to generate the message type.
        cls._meta.generate_pb()

        # Iterate over each field, if the field has any messages which
        # have not been instantiated, register them against the MessageRegistry
        # to eventually save the instantiated type.
        for field in cls._meta.fields.values():
            if not field.ready:
                registry.expect(
                    field=field,
                    message_name=field.message
                        if isinstance(field.message, str)
                        else field.message.__qualname__,
                    package=package,
                )

        # Done; return the message class.
        return cls

    @classmethod
    def __prepare__(mcls, name, bases, **kwargs):
        return collections.OrderedDict()

    @property
    def meta(cls):
        return cls._meta

    def pb(cls, obj=None):
        """Return the underlying protobuf Message class or instance.

        Args:
            obj: If provided, and an instance of ``cls``, return the
                underlying protobuf instance.
        """
        if obj is None:
            return cls.meta.pb
        if not isinstance(obj, cls):
            raise TypeError('%r is not an instance of %s' % (
                obj, cls.__name__,
            ))
        return obj._pb

    def wrap(cls, pb):
        """Return a Message object that shallowly wraps the descriptor.

        Args:
            pb: A protocol buffer object, such as would be returned by
                :meth:`pb`.
        """
        return cls(pb, __wrap_original=True)

    def serialize(cls, instance) -> bytes:
        """Return the serialized proto.

        Args:
            instance: An instance of this message type.

        Returns:
            bytes: The serialized representation of the protocol buffer.
        """
        return cls.pb(instance).SerializeToString()

    def deserialize(cls, payload: bytes) -> 'Message':
        """Given a serialized proto, deserialize it into a Message instance.

        Args:
            payload (bytes): The serialized proto.

        Returns:
            ~.Message: An instance of the message class against which this
            method was called.
        """
        return cls(cls.pb().FromString(payload))


class Message(metaclass=MessageMeta):
    """The abstract base class for a message.

    Args:
        mapping (Union[dict, ~.Message]): A dictionary or message to be
            used to determine the values for this message.
        kwargs (dict): Keys and values corresponding to the fields of the
            message.
    """
    def __init__(self, mapping=None, **kwargs):
        # We accept several things for `mapping`:
        #   * An instance of this class.
        #   * An instance of the underlying protobuf descriptor class.
        #   * A dict
        #   * Nothing (keyword arguments only).
        #
        # Sanity check: Did we get something not on that list? Error if so.
        if mapping and not isinstance(
                mapping, (collections.abc.Mapping, type(self), self._meta.pb)):
            raise TypeError('Invalid constructor input for %s: %r' % (
                self.__class__.__name__, mapping,
            ))

        # Handle the first two cases: they both involve keeping
        # a copy of the underlying protobuf descriptor instance.
        if isinstance(mapping, type(self)):
            mapping = mapping._pb
        if isinstance(mapping, self._meta.pb):
            # Make a copy of the mapping.
            # This is a constructor for a new object, so users will assume
            # that it will not have side effects on the arguments being
            # passed in.
            #
            # The `__wrap_original` argument is private API to override
            # this behavior, because `MessageMarshal` actually does want to
            # wrap the original argument it was given. The `wrap` method
            # on the metaclass is the public API for this behavior.
            if not kwargs.pop('__wrap_original', False):
                mapping = copy.copy(mapping)
            self._pb = mapping
            if kwargs:
                self._pb.MergeFrom(self._meta.pb(**kwargs))
            return

        # Handle the remaining case by converging the mapping and kwargs
        # dictionaries (with kwargs winning), and saving a descriptor
        # based on that.
        if mapping is None:
            mapping = {}
        mapping.update(kwargs)

        # Update the mapping to address any values that need to be
        # coerced.
        for key, value in copy.copy(mapping).items():
            pb_type = self._meta.fields[key].pb_type
            pb_value = marshal.to_proto(pb_type, value)
            if pb_value is None:
                mapping.pop(key)
            else:
                mapping[key] = pb_value

        # Create the internal protocol buffer.
        self._pb = self._meta.pb(**mapping)

    def __bool__(self):
        """Return True if any field is truthy, False otherwise."""
        return any([k in self and getattr(self, k)
                    for k in self._meta.fields.keys()])

    def __contains__(self, key):
        """Return True if this field was set to something non-zero on the wire.

        In most cases, this method will return True when ``__getattr__``
        would return a truthy value and False when it would return a falsy
        value, so explicitly calling this is not useful.

        The exception case is empty messages explicitly set on the wire,
        which are falsy from ``__getattr__``. This method allows to
        distinguish between an explicitly provided empty message and the
        absence of that message, which is useful in some edge cases.

        The most common edge case is the use of ``google.protobuf.BoolValue``
        to get a boolean that distinguishes between ``False`` and ``None``
        (or the same for a string, int, etc.). This library transparently
        handles that case for you, but this method remains available to
        accomodate cases not automatically covered.

        Args:
            key (str): The name of the field.

        Returns:
            bool: Whether the field's value corresponds to a non-empty
                wire serialization.
        """
        pb_value = getattr(self._pb, key)
        try:
            # Protocol buffers "HasField" is unfriendly; it only works
            # against composite, non-repeated fields, and raises ValueError
            # against any repeated field or primitive.
            #
            # There is no good way to test whether it is valid to provide
            # a field to this method, so sadly we are stuck with a
            # somewhat inefficient try/except.
            return self._pb.HasField(key)
        except ValueError:
            return bool(pb_value)

    def __delattr__(self, key):
        """Delete the value on the given field.

        This is generally equivalent to setting a falsy value.
        """
        self._pb.ClearField(key)

    def __eq__(self, other):
        """Return True if the messages are equal, False otherwise."""
        # If these are the same type, use internal protobuf's equality check.
        if isinstance(other, type(self)):
            return self._pb == other._pb

        # If the other type is the target protobuf object, honor that also.
        if isinstance(other, self._meta.pb):
            return self._pb == other

        # Ask the other object.
        return NotImplemented

    def __getattr__(self, key):
        """Retrieve the given field's value.

        In protocol buffers, the presence of a field on a message is
        sufficient for it to always be "present".

        For primitives, a value of the correct type will always be returned
        (the "falsy" values in protocol buffers consistently match those
        in Python). For repeated fields, the falsy value is always an empty
        sequence.

        For messages, protocol buffers does distinguish between an empty
        message and absence, but this distinction is subtle and rarely
        relevant. Therefore, this method always returns an empty message
        (following the official implementation). To check for message
        presence, use ``key in self`` (in other words, ``__contains__``).

        .. note::

            Some well-known protocol buffer types
            (e.g. ``google.protobuf.Timestamp``) will be converted to
            their Python equivalents. See the ``marshal`` module for
            mode details.
        """
        pb_type = self._meta.fields[key].pb_type
        pb_value = getattr(self._pb, key)
        return marshal.to_python(pb_type, pb_value, absent=key not in self)

    def __ne__(self, other):
        """Return True if the messages are unequal, False otherwise."""
        return not self == other

    def __repr__(self):
        return repr(self._pb)

    def __setattr__(self, key, value):
        """Set the value on the given field.

        For well-known protocol buffer types which are marshalled, either
        the protocol buffer object or the Python equivalent is accepted.
        """
        if key.startswith('_'):
            return super().__setattr__(key, value)
        pb_type = self._meta.fields[key].pb_type
        pb_value = marshal.to_proto(pb_type, value)

        # We *always* clear the existing field.
        # This is the only way to successfully write nested falsy values,
        # because otherwise MergeFrom will no-op on them.
        self._pb.ClearField(key)

        # Merge in the value being set.
        if pb_value is not None:
            self._pb.MergeFrom(self._meta.pb(**{key: pb_value}))


class MessageInfo:
    """Metadata about a message.

    Args:
        pb (type): The underlying protobuf message.
        fields (Tuple[~.fields.Field]): The fields declared on the message.
        package (str): The proto package
        full_name (str): The full name of the message.
        options (~.descriptor_pb2.MessageOptions): Any options that were
            set on the message.
    """
    def __init__(self, *, descriptor: descriptor.Descriptor,
                 fields: List[Field],
                 package: str, full_name: str,
                 options: MessageOptions) -> None:
        self.descriptor = descriptor
        self.package = package
        self.full_name = full_name
        self.options = options
        self.fields = collections.OrderedDict([
            (i.name, i) for i in fields
        ])
        self.fields_by_number = collections.OrderedDict([
            (i.number, i) for i in fields
        ])
        self._pb = None

    @property
    def pb(self) -> Type[message.Message]:
        """Return the protobuf message type for this descriptor.

        If a field on the message references another message which has not
        loaded, then this method returns None.
        """
        return self._pb

    def generate_pb(self):
        """Return the protobuf message type for this descriptor.

        If a field on the message references another message which has not
        loaded, then this method returns None.

        Once this method attaches a value once, the return value is
        cached and returned from subsequent calls.
        """
        desc = self.descriptor

        # Corner case: Self-referential messages.
        # In this case, assign the newly-created message up front.
        for field in self.fields.values():
            if field.message == self.parent.__qualname__:
                field.message = self.parent
                desc.fields_by_name[field.name].message_type = desc

        # If this message is not ready, hold off.
        if not self.ready:
            return

        # Only make a new descriptor if we do not already have one.
        if not self._pb:
            # Create the stock protobuf Message.
            pb_message = reflection.GeneratedProtocolMessageType(
                self.full_name.split('.')[-1], (message.Message,), {
                    'DESCRIPTOR': desc,
                    '__module__': None,
                },
            )
            symbol_database.Default().RegisterMessage(pb_message)

            # Save the message.
            self._pb = pb_message

            # Register this class with the message registry.
            # This handles forward references in the event that a message needs
            # to reference a message defined later in the file.
            registry.register(self.parent)

            # Register the new class with the marshal.
            marshal.register(
                pb_message,
                MessageMarshal(pb_message, self.parent),
            )

    @property
    def ready(self):
        return all([i.ready for i in self.fields.values()])


class MessageRegistry:
    """An overall registry of messages.

    Protocol buffers allows referencing messages earlier in the file than
    they are declared, and we wish to preserve declaration order.

    This registry allows fields to register themselves as placeholders,
    and they are given message references once those messages are
    instantiated.
    """
    def __init__(self):
        self._registry = {}
        self._expecting = {}

    def clear(self):
        """Clear the registry."""
        self._registry.clear()
        self._expecting.clear()

    def expect(self, *, package: str, message_name: str, field: Field):
        """Register a field as expecting a message.

        Args:
            package (str): The proto package (or empty string if there is no
                proto package).
            message_name (str): The name of the message.
            field (~.Field): The field which should reference the message.
        """
        key = (package, message_name)

        # Sanity check: If we already instantiated this message, then
        # fulfill the expectation immediately.
        if key in self._registry:
            self._attach(field, self._registry[key])

        # Store that this field is expecting instantiation of this message.
        self._expecting.setdefault(key, [])
        self._expecting[key].append(field)

    def register(self, message: Type[Message]):
        """Register a message with the registry."""
        key = (message._meta.package, message.__qualname__)
        self._registry[key] = message

        # If any fields are expecting this message, attach the message
        # to them.
        for field in self._expecting.pop(key, ()):
            self._attach(field, message)

        # A expects B
        # B is instantiated, but expects C
        # C is instantiated, fulfills B
        #   ...which must notify A.

    def _attach(self, field, message):
        # Add the instantiated message to the field.
        field.message = message
        field.descriptor.message_type = message._meta.pb.DESCRIPTOR

        # The field's containing message *may* (or may not) now have all of
        # its fields become concrete. Attempt to populate its message type.
        field.parent._meta.generate_pb()


registry = MessageRegistry()


_file_descriptor_registry = {}


__all__ = (
    'Message',
    'registry',
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum

import proto


def test_outer_enum_init():
    class Color(enum.IntEnum):
        COLOR_UNSPECIFIED = 0
        RED = 1
        GREEN = 2
        BLUE = 3

    class Foo(proto.Message):
        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo(color=Color.RED)
    assert foo.color == Color.RED
    assert foo.color == 1
    assert foo.color
    assert Foo.pb(foo).color == 1


def test_outer_enum_init_int():
    class Color(enum.IntEnum):
        COLOR_UNSPECIFIED = 0
        RED = 1
        GREEN = 2
        BLUE = 3

    class Foo(proto.Message):
        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo(color=1)
    assert foo.color == Color.RED
    assert foo.color == 1
    assert foo.color
    assert Foo.pb(foo).color == 1


def test_outer_enum_init_dict():
    class Color(enum.IntEnum):
        COLOR_UNSPECIFIED = 0
        RED = 1
        GREEN = 2
        BLUE = 3

    class Foo(proto.Message):
        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo({'color': 1})
    assert foo.color == Color.RED
    assert foo.color == 1
    assert foo.color
    assert Foo.pb(foo).color == 1


def test_outer_enum_init_pb2():
    class Color(enum.IntEnum):
        COLOR_UNSPECIFIED = 0
        RED = 1
        GREEN = 2
        BLUE = 3

    class Foo(proto.Message):
        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo(Foo.pb()(color=Color.RED))
    assert foo.color == Color.RED
    assert foo.color == 1
    assert foo.color
    assert Foo.pb(foo).color == 1


def test_outer_enum_unset():
    class Color(enum.IntEnum):
        COLOR_UNSPECIFIED = 0
        RED = 1
        GREEN = 2
        BLUE = 3

    class Foo(proto.Message):
        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo()
    assert foo.color == Color.COLOR_UNSPECIFIED
    assert foo.color == 0
    assert 'color' not in foo
    assert not foo.color
    assert Foo.pb(foo).color == 0
    assert Foo.serialize(foo) == b''


def test_outer_enum_write():
    class Color(enum.IntEnum):
        COLOR_UNSPECIFIED = 0
        RED = 1
        GREEN = 2
        BLUE = 3

    class Foo(proto.Message):
        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo()
    foo.color = Color.GREEN
    assert foo.color == Color.GREEN
    assert foo.color == 2
    assert Foo.pb(foo).color == 2
    assert foo.color


def test_outer_enum_write_int():
    class Color(enum.IntEnum):
        COLOR_UNSPECIFIED = 0
        RED = 1
        GREEN = 2
        BLUE = 3

    class Foo(proto.Message):
        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo()
    foo.color = 3
    assert foo.color == Color.BLUE
    assert foo.color == 3
    assert Foo.pb(foo).color == 3
    assert foo.color


def test_inner_enum_init():
    class Foo(proto.Message):
        class Color(enum.IntEnum):
            COLOR_UNSPECIFIED = 0
            RED = 1
            GREEN = 2
            BLUE = 3

        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo(color=Foo.Color.RED)
    assert foo.color == Foo.Color.RED
    assert foo.color == 1
    assert foo.color
    assert Foo.pb(foo).color == 1


def test_inner_enum_write():
    class Foo(proto.Message):
        class Color(enum.IntEnum):
            COLOR_UNSPECIFIED = 0
            RED = 1
            GREEN = 2
            BLUE = 3

        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo()
    foo.color = Foo.Color.GREEN
    assert foo.color == Foo.Color.GREEN
    assert foo.color == 2
    assert Foo.pb(foo).color == 2
    assert foo.color


def test_enum_del():
    class Color(enum.IntEnum):
        COLOR_UNSPECIFIED = 0
        RED = 1
        GREEN = 2
        BLUE = 3

    class Foo(proto.Message):
        color = proto.Field(proto.ENUM, number=1, enum=Color)

    foo = Foo(color=Color.BLUE)
    del foo.color
    assert foo.color == Color.COLOR_UNSPECIFIED
    assert foo.color == 0
    assert 'color' not in foo
    assert not foo.color
    assert Foo.pb(foo).color == 0



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto


def test_composite_map():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.MapField(proto.STRING, proto.MESSAGE, number=1,
            message=Foo,
        )

    baz = Baz(foos={'i': Foo(bar=42), 'j': Foo(bar=24)})
    assert len(baz.foos) == 2
    assert baz.foos['i'].bar == 42
    assert baz.foos['j'].bar == 24
    assert 'k' not in baz.foos


def test_composite_map_dict():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.MapField(proto.STRING, proto.MESSAGE, number=1,
            message=Foo,
        )

    baz = Baz(foos={'i': {'bar': 42}, 'j': {'bar': 24}})
    assert len(baz.foos) == 2
    assert baz.foos['i'].bar == 42
    assert baz.foos['j'].bar == 24
    assert 'k' not in baz.foos


def test_composite_map_set():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.MapField(proto.STRING, proto.MESSAGE, number=1,
            message=Foo,
        )

    baz = Baz()
    baz.foos['i'] = Foo(bar=42)
    baz.foos['j'] = Foo(bar=24)
    assert len(baz.foos) == 2
    assert baz.foos['i'].bar == 42
    assert baz.foos['j'].bar == 24
    assert 'k' not in baz.foos


def test_composite_map_deep_set():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.MapField(proto.STRING, proto.MESSAGE, number=1,
            message=Foo,
        )

    baz = Baz()
    baz.foos['i'] = Foo()
    baz.foos['i'].bar = 42
    assert len(baz.foos) == 1
    assert baz.foos['i'].bar == 42


def test_composite_map_del():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Baz(proto.Message):
        foos = proto.MapField(proto.STRING, proto.MESSAGE, number=1,
            message=Foo,
        )

    baz = Baz()
    baz.foos['i'] = Foo(bar=42)
    assert len(baz.foos) == 1
    del baz.foos['i']
    assert len(baz.foos) == 0
    assert 'i' not in baz.foos



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import proto


def test_message_constructor_instance():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)
    foo_original = Foo(bar=42)
    foo_copy = Foo(foo_original)
    assert foo_original.bar == foo_copy.bar == 42
    assert foo_original == foo_copy
    assert foo_original is not foo_copy
    assert isinstance(foo_original, Foo)
    assert isinstance(foo_copy, Foo)
    assert isinstance(Foo.pb(foo_copy), Foo.pb())


def test_message_constructor_underlying_pb2():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)
    foo_pb2 = Foo.pb()(bar=42)
    foo = Foo(foo_pb2)
    assert foo.bar == Foo.pb(foo).bar == foo_pb2.bar == 42
    assert foo == foo_pb2  # Not communitive. Nothing we can do about that.
    assert foo_pb2 == Foo.pb(foo)
    assert foo_pb2 is not Foo.pb(foo)
    assert isinstance(foo, Foo)
    assert isinstance(Foo.pb(foo), Foo.pb())
    assert isinstance(foo_pb2, Foo.pb())


def test_message_constructor_underlying_pb2_and_kwargs():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)
    foo_pb2 = Foo.pb()(bar=42)
    foo = Foo(foo_pb2, bar=99)
    assert foo.bar == Foo.pb(foo).bar == 99
    assert foo_pb2.bar == 42
    assert isinstance(foo, Foo)
    assert isinstance(Foo.pb(foo), Foo.pb())
    assert isinstance(foo_pb2, Foo.pb())


def test_message_constructor_dict():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)
    foo = Foo({'bar': 42})
    assert foo.bar == Foo.pb(foo).bar == 42
    assert foo != {'bar': 42}
    assert isinstance(foo, Foo)
    assert isinstance(Foo.pb(foo), Foo.pb())


def test_message_constructor_kwargs():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)
    foo = Foo(bar=42)
    assert foo.bar == Foo.pb(foo).bar == 42
    assert isinstance(foo, Foo)
    assert isinstance(Foo.pb(foo), Foo.pb())


def test_message_constructor_invalid():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)
    with pytest.raises(TypeError):
        Foo(object())


def test_message_contains_primitive():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)
    assert 'bar' in Foo(bar=42)
    assert 'bar' not in Foo(bar=0)
    assert 'bar' not in Foo()


def test_message_contains_composite():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)

    class Baz(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message=Foo)

    assert 'foo' in Baz(foo=Foo(bar=42))
    assert 'foo' in Baz(foo=Foo())
    assert 'foo' not in Baz()


def test_message_contains_repeated_primitive():
    class Foo(proto.Message):
        bar = proto.RepeatedField(proto.INT64, number=1)

    assert 'bar' in Foo(bar=[1, 1, 2, 3, 5])
    assert 'bar' in Foo(bar=[0])
    assert 'bar' not in Foo(bar=[])
    assert 'bar' not in Foo()


def test_message_contains_repeated_composite():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)

    class Baz(proto.Message):
        foo = proto.RepeatedField(proto.MESSAGE, number=1, message=Foo)

    assert 'foo' in Baz(foo=[Foo(bar=42)])
    assert 'foo' in Baz(foo=[Foo()])
    assert 'foo' not in Baz(foo=[])
    assert 'foo' not in Baz()


def test_message_eq_primitives():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)
        baz = proto.Field(proto.STRING, number=2)
        bacon = proto.Field(proto.BOOL, number=3)

    assert Foo() == Foo()
    assert Foo(bar=42, baz='42') == Foo(bar=42, baz='42')
    assert Foo(bar=42, baz='42') != Foo(baz='42')
    assert Foo(bar=42, bacon=True) == Foo(bar=42, bacon=True)
    assert Foo(bar=42, bacon=True) != Foo(bar=42)
    assert Foo(bar=42, baz='42', bacon=True) != Foo(bar=42, bacon=True)
    assert Foo(bacon=False) == Foo()
    assert Foo(bacon=True) != Foo(bacon=False)
    assert Foo(bar=21 * 2) == Foo(bar=42)
    assert Foo() == Foo(bar=0)
    assert Foo() == Foo(bar=0, baz='', bacon=False)
    assert Foo() != Foo(bar=0, baz='0', bacon=False)


def test_message_serialize():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)
        baz = proto.Field(proto.STRING, number=2)
        bacon = proto.Field(proto.BOOL, number=3)

    foo = Foo(bar=42, bacon=True)
    assert Foo.serialize(foo) == Foo.pb(foo).SerializeToString()


def test_message_deserialize():
    class OldFoo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class NewFoo(proto.Message):
        bar = proto.Field(proto.INT64, number=1)

    serialized = OldFoo.serialize(OldFoo(bar=42))
    new_foo = NewFoo.deserialize(serialized)
    assert isinstance(new_foo, NewFoo)
    assert new_foo.bar == 42


def test_message_generate_pb_idempotent():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    Foo._meta.generate_pb()
    pb = Foo.pb()
    Foo._meta.generate_pb()
    assert pb is Foo.pb()


def test_message_pb():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    assert isinstance(Foo.pb(Foo()), Foo.pb())
    with pytest.raises(TypeError):
        Foo.pb(object())



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto


def test_composite_forward_ref():
    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message='Foo')
        eggs = proto.Field(proto.BOOL, number=2)

    class Foo(proto.Message):
        __qualname__ = 'Foo'
        bar = proto.Field(proto.STRING, number=1)
        baz = proto.Field(proto.INT64, number=2)

    spam = Spam(foo=Foo(bar='str', baz=42))
    assert spam.foo.bar == 'str'
    assert spam.foo.baz == 42
    assert spam.eggs is False


def test_composite_backward_ref():
    class Foo(proto.Message):
        __qualname__ = 'Foo'
        bar = proto.Field(proto.STRING, number=1)
        baz = proto.Field(proto.INT64, number=2)

    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message='Foo')
        eggs = proto.Field(proto.BOOL, number=2)

    spam = Spam(foo=Foo(bar='str', baz=42))
    assert spam.foo.bar == 'str'
    assert spam.foo.baz == 42
    assert spam.eggs is False


def test_composite_multi_ref():
    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message='Foo')
        eggs = proto.Field(proto.BOOL, number=2)

    class Foo(proto.Message):
        __qualname__ = 'Foo'
        bar = proto.Field(proto.STRING, number=1)
        baz = proto.Field(proto.INT64, number=2)

    class Bacon(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message='Foo')

    spam = Spam(foo=Foo(bar='str', baz=42))
    bacon = Bacon(foo=spam.foo)
    assert spam.foo.bar == 'str'
    assert spam.foo.baz == 42
    assert spam.eggs is False
    assert bacon.foo == spam.foo


def test_composite_self_ref():
    class Spam(proto.Message):
        __qualname__ = 'Spam'
        spam = proto.Field(proto.MESSAGE, number=1, message='Spam')
        eggs = proto.Field(proto.BOOL, number=2)

    spam = Spam(spam=Spam(eggs=True))
    assert spam.eggs is False
    assert spam.spam.eggs is True
    assert type(spam) is type(spam.spam)  # noqa: E0721
    assert not spam.spam.spam
    assert spam.spam.spam.eggs is False
    assert not spam.spam.spam.spam.spam.spam.spam.spam



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto


def test_string_init():
    class Foo(proto.Message):
        bar = proto.Field(proto.STRING, number=1)
        baz = proto.Field(proto.STRING, number=2)

    foo = Foo(bar='spam')
    assert foo.bar == 'spam'
    assert foo.baz == ''
    assert not foo.baz
    assert Foo.pb(foo).bar == 'spam'
    assert Foo.pb(foo).baz == ''


def test_string_rmw():
    class Foo(proto.Message):
        spam = proto.Field(proto.STRING, number=1)
        eggs = proto.Field(proto.STRING, number=2)

    foo = Foo(spam='bar')
    foo.eggs = 'baz'
    assert foo.spam == 'bar'
    assert foo.eggs == 'baz'
    assert Foo.pb(foo).spam == 'bar'
    assert Foo.pb(foo).eggs == 'baz'
    foo.spam = 'bacon'
    assert foo.spam == 'bacon'
    assert foo.eggs == 'baz'
    assert Foo.pb(foo).spam == 'bacon'
    assert Foo.pb(foo).eggs == 'baz'


def test_string_del():
    class Foo(proto.Message):
        bar = proto.Field(proto.STRING, number=1)

    foo = Foo(bar='spam')
    assert foo.bar == 'spam'
    del foo.bar
    assert foo.bar == ''
    assert not foo.bar



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os

from setuptools import find_packages, setup


PACKAGE_ROOT = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(PACKAGE_ROOT, 'README.rst')) as file_obj:
    README = file_obj.read()

setup(
    name='proto-plus',
    version='0.1.0a4',
    license='Apache 2.0',
    author='Luke Sneeringer',
    author_email='lukesneeringer@google.com',
    url='https://github.com/googleapis/proto-plus-python.git',
    packages=find_packages(exclude=['docs', 'tests']),
    namespace_packages=['google', 'google.protobuf'],
    description='Beautiful, Pythonic protocol buffers.',
    long_description=README,
    platforms='Posix; MacOS X',
    include_package_data=True,
    install_requires=(
        'protobuf >= 3.5.1',
    ),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    zip_safe=False,
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from google.protobuf import descriptor_pb2

import proto


def test_registration():
    FileDescriptor = descriptor_pb2.FileDescriptorProto
    try:
        @proto.marshal.register(FileDescriptor)
        class Marshal:
            def to_proto(self, value):
                return value

            def to_python(self, value, *, absent=None):
                return value
        assert isinstance(proto.marshal._registry[FileDescriptor], Marshal)
    finally:
        proto.marshal._registry.pop(FileDescriptor, None)


def test_invalid_marshal_class():
    with pytest.raises(TypeError):
        @proto.marshal.register(descriptor_pb2.FileDescriptorProto)
        class Marshal:
            pass


def test_invalid_marshal_rule():
    with pytest.raises(TypeError):
        proto.marshal.register(descriptor_pb2.DescriptorProto, rule=object())



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.protobuf import descriptor
from google.protobuf import message as pb_message

from proto.primitives import get_default_value
from proto.primitives import ProtoType


class Field:
    """A representation of a type of field in protocol buffers."""

    def __init__(self, proto_type, *, number: int,
                 message=None, enum=None, oneof: str = None):
        # This class is not intended to stand entirely alone;
        # data is augmented by the metaclass for Message.
        self.mcls_data = {}
        self.parent = None

        # Save the direct arguments.
        self.number = number
        self.proto_type = proto_type
        self.message = message
        self.enum = enum
        self.oneof = oneof

        # Fields are neither repeated nor maps.
        # The RepeatedField and MapField subclasses override these values
        # in their initializers.
        self.repeated = False

        # Once the descriptor is accessed the first time, cache it.
        # This is important because in rare cases the message or enum
        # types are written later.
        self._descriptor = None

    @property
    def descriptor(self):
        """Return the descriptor for the field."""
        if not self._descriptor:
            # Determine the default value.
            default_value = get_default_value(self.proto_type)
            if self.repeated:
                default_value = []

            # Resolve the message type, if any, to its underlying
            # protobuf descriptor.
            message_type = self.message
            if self.message and hasattr(self.message, '_meta'):
                message_type = self.message._meta.pb
            elif isinstance(self.message, str):
                # If the message was specified as a string, skip passing it to
                # the descriptor for the moment (when the field is added to a
                # message in `MessageMeta`, it will be resolved).
                message_type = None

            # Set the descriptor.
            self._descriptor = descriptor.FieldDescriptor(
                name=self.mcls_data['name'],
                full_name=self.mcls_data['full_name'],
                index=self.mcls_data['index'],
                number=self.number,
                label=3 if self.repeated else 1,
                type=self.proto_type,
                cpp_type=descriptor.FieldDescriptor.ProtoTypeToCppProtoType(
                    self.proto_type,
                ),
                default_value=default_value,
                message_type=message_type.DESCRIPTOR if message_type else None,
                enum_type=self.enum,
                containing_type=None,
                is_extension=False,
                extension_scope=None,
            )
        return self._descriptor

    @property
    def name(self):
        """Return the name of the field."""
        return self.mcls_data['name']

    @property
    def pb_type(self):
        if self.message:
            if hasattr(self.message, '_meta'):
                return self.message.pb()
            return self.message
        return None

    @property
    def ready(self):
        # All non-message fields are ready.
        # We do not have to check readiness on enums because we process
        # all enums ahead of all messages, thus guaranteeing that they must
        # be processed first.
        if self.proto_type != ProtoType.MESSAGE:
            return True

        # This field may rely on an instantiated stock protobuf message.
        # These are ready.
        if (isinstance(self.message, type) and
                issubclass(self.message, pb_message.Message)):
            return True

        # Corner case: Self-referential messages.
        # Explicitly declare these to be ready.
        if self.parent is self.message:
            return True

        # This field is ready if the underlying message is.
        # That means the message is a message object (not a string) and that
        # the referenced message is also ready.
        # We use the `.pb` property rather than `.ready` here to avoid
        # infinite recursion corner cases (e.g. self-referencing messages).
        if isinstance(self.message, str):
            return False
        return bool(self.message._meta.pb)


class RepeatedField(Field):
    """A representation of a repeated field in protocol buffers."""

    def __init__(self, proto_type, *, number: int,
                 message=None, enum=None):
        super().__init__(proto_type, number=number, message=message, enum=enum)
        self.repeated = True


class MapField(Field):
    """A representation of a map field in protocol buffers."""

    def __init__(self, key_type, value_type, *, number: int,
                 message=None, enum=None):
        super().__init__(value_type, number=number, message=message, enum=enum)
        self.map_key_type = key_type


__all__ = (
    'Field',
    'MapField',
    'RepeatedField',
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum


class ProtoType(enum.IntEnum):
    """The set of basic types in protocol buffers."""
    # These values come from google/protobuf/descriptor.proto
    DOUBLE = 1
    FLOAT = 2
    INT64 = 3
    UINT64 = 4
    INT32 = 5
    FIXED64 = 6
    FIXED32 = 7
    BOOL = 8
    STRING = 9
    MESSAGE = 11
    BYTES = 12
    UINT32 = 13
    ENUM = 14
    SFIXED32 = 15
    SFIXED64 = 16
    SINT32 = 17
    SINT64 = 18


# The default values for each type.
def get_default_value(proto_type):
    return {
        ProtoType.DOUBLE: 0.0,
        ProtoType.FLOAT: 0.0,
        ProtoType.INT64: 0,
        ProtoType.UINT64: 0,
        ProtoType.INT32: 0,
        ProtoType.FIXED64: 0,
        ProtoType.FIXED32: 0,
        ProtoType.BOOL: False,
        ProtoType.STRING: '',
        ProtoType.MESSAGE: None,
        ProtoType.BYTES: b'',
        ProtoType.UINT32: 0,
        ProtoType.ENUM: 0,
        ProtoType.SFIXED32: 0,
        ProtoType.SFIXED64: 0,
        ProtoType.SINT32: 0,
        ProtoType.SINT64: 0,
    }[proto_type]



# Force native protocol buffers into the Python (not C) implementation.
import os
os.environ['PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION'] = 'python'

# Declare the usual Python namespace packaging jig.
if __name__ != '__main__':
    try:
        __import__('pkg_resources').declare_namespace(__name__)
    except ImportError:
        __path__ = __import__('pkgutil').extend_path(__path__, __name__)
